#ifdef __USAGE
%C - SCSI utility

%C [-n node] [[device] command]*
 scan       Scan for all SCSI devices in /dev
 lock       Lock the device
 unlock     Unlock the device
 start      Start unit
 stop       Stop unit
 eject      Eject the device
 error      Detail on last error device received
 inquiry    Do a SCSI inquiry to the device
 info       Report SCSI info on the device
 sense      Display Mode sense page
 defect     Display defect list
 density    Display or change density
 medium     Display or change medium

device can be a full path to the device, or basename
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <sys/vc.h>
#include <sys/qioctl.h>
#include <sys/sendmx.h>
#include <sys/uscsi.h>
#include "scsi.h"
#include "scsi_sense.h"

nid_t				nid = 0;
int					fd = -1;
char				path[PATH_MAX];
int					oflag;

struct uscsi_cmd	scsicmd;
char				scsicdb[12];
char				scsibuf[350];

int uscsi(int fd, unsigned request, struct uscsi_cmd *cmd) {
	struct _mxfer_entry		mx[3];
	
	_setmx(&mx[0], cmd, sizeof *cmd);
	_setmx(&mx[1], cmd->uscsi_cdb, cmd->uscsi_cdblen);
	_setmx(&mx[2], cmd->uscsi_bufaddr, cmd->uscsi_buflen);

	return qnx_ioctlmx(fd, request, 3, 3, mx, mx);
}

int scsi_cmd(int error, char *desc) {
	int						ret;
	struct scsi_cmds		*c;
	
	printf("Executing Command      : %s ", desc);
	fflush(stdout);
	ret = uscsi(fd, USCSICMD, &scsicmd);
	if(ret == -1 || scsicmd.uscsi_status != USCSI_STATUS_GOOD) {
		printf("(Failed)\n");
		if(error) {
			if(ret == -1) {
				printf("SCSI Command Not Sent  : %02Xh", scsicmd.uscsi_cdb[0]);
			} else {
				printf("SCSI Command Failed    : %02Xh", scsicmd.uscsi_cdb[0]); 
			}
			for(c = scsi_cmds; c->desc; c++) {
				if(c->cmd == scsicmd.uscsi_cdb[0]) {
					break;
				}
			}
			if(c->desc) {
				printf(" (%s)", c->desc);
			} else {
				printf(" (unknown cmd)");
			}
		} else {
			return -1;
		}
	} else {
		printf("(Succeded)\n");
	}
	if(error) {
		if(ret == -1) {
			printf(" (%d-%s)\n", errno, strerror(errno));
		} else {
			if(scsicmd.uscsi_status != USCSI_STATUS_GOOD) {
				printf(" (Scsi status=%02Xh)\n", scsicmd.uscsi_status);
				if(scsicmd.uscsi_status & USCSI_STATUS_CHECK) {
					int					cmd_error(void);
	
					switch(cmd_error()) {
					case USCSI_SK_NO_SENSE:
					case USCSI_SK_RECOVERED:
						ret = 0;
						break;
					default:
						ret = -1;
						break;
					}
				}
			}
		}
	}
	return ret;
}

void cmd_start(void) {
    scsicmd.uscsi_flags = USCSI_NOFLUSH;
	scsicdb[0] = SCSI_OP(0, SC_REMOVAL);
	scsicdb[4] = 0x01;
	scsicmd.uscsi_cdblen = 6;
    scsicmd.uscsi_buflen = 0;
	scsi_cmd(1, "start unit");
}

void cmd_stop(void) {
    scsicmd.uscsi_flags = USCSI_NOFLUSH;
	scsicdb[0] = SCSI_OP(0, SC_LOAD);
	scsicmd.uscsi_cdblen = 6;
    scsicmd.uscsi_buflen = 0;
	scsi_cmd(1, "stop unit");
}

void cmd_lock(void) {
    scsicmd.uscsi_flags = USCSI_NOFLUSH;
	scsicdb[0] = SCSI_OP(0, SC_REMOVAL);
	scsicdb[4] = 0x01;
	scsicmd.uscsi_cdblen = 6;
    scsicmd.uscsi_buflen = 0;
	scsi_cmd(1, "lock");
}

void cmd_unlock(void) {
    scsicmd.uscsi_flags = USCSI_NOFLUSH;
	scsicdb[0] = SCSI_OP(0, SC_REMOVAL);
	scsicmd.uscsi_cdblen = 6;
    scsicmd.uscsi_buflen = 0;
	scsi_cmd(1, "unlock");
}

void cmd_unload(void) {
	struct uscsi_info		info;

	if(qnx_ioctl(fd, USCSIINFO, 0, 0, &info, sizeof info) == -1) {
		printf("Couldn't device type information\n");
		return;
	}
	switch(info.peripheral) {
	case D_SEQ_ACC:
		scsicdb[0] = SCSI_OP(0, SC_LOAD);
		scsicdb[4] = 0x00;
		break;
	default:
		scsicdb[0] = SCSI_OP(0, SC_SPINDLE);
		scsicdb[4] = 0x02;
	}
	scsicmd.uscsi_cdblen = 6;
    scsicmd.uscsi_buflen = 0;
	scsi_cmd(1, "eject");
}

void cmd_load(void) {
	struct uscsi_info		info;

	if(qnx_ioctl(fd, USCSIINFO, 0, 0, &info, sizeof info) == -1) {
		printf("Couldn't device type information\n");
		return;
	}
	switch(info.peripheral) {
	case D_SEQ_ACC:
		scsicdb[0] = SCSI_OP(0, SC_LOAD);
		scsicdb[4] = 0x01;
		break;
	default:
		scsicdb[0] = SCSI_OP(0, SC_SPINDLE);
		scsicdb[4] = 0x03;
	}
	scsicmd.uscsi_cdblen = 6;
    scsicmd.uscsi_buflen = 0;
	scsi_cmd(1, "eject");
}

int cmd_error(void) {
	struct uscsi_sense		sense;
	long					n;
	unsigned char			*p;
	struct scsi_asc			*s1, *s2;
	int						sk = -1;

	if(qnx_ioctl(fd, USCSISENSE, 0, 0, &sense, sizeof sense) == -1) {
		printf("Couldn't get sense information\n");
		return sk;
	}
	if((sense.sd_err & 0x70) != 0x70) {
		printf("Error Class            : %d (Nonextended sense data)\n", (sense.sd_err >> 4) & 0x07);
		printf("Error Code             : %d\n", sense.sd_err & 0x0f);
		p = (unsigned char *)&sense.sd_seg;
		printf("Vender Unique          : %d\n", p[0] >> 4);
		if(sense.sd_err & 0x80) {
			n = (((long)p[0] & 0x0f) << 16) | ((long)p[1] << 8) | (long)p[2];
			printf("Logical Block Address  : %lxh or %ld\n", n, n);
		}
		return sk;
	}
	printf("Error Class            : %d (Extended sense data)\n", (sense.sd_err >> 4) & 0x07);
	printf("Error Code             : %d", sense.sd_err & 0x0f);
	switch(sense.sd_err & 0x0f) {
	case 0x00:
		printf(" (current error)\n");
		sk = sense.sd_sense & SK_MASK;
		break;
	case 0x01:
		printf(" (deferred error)\n");
		break;
	case 0x0f:
		printf(" (vender unique)\n");
		break;
	default:
		printf(" (reserved)\n");
		break;
	}
	if(sense.sd_seg) printf("Segment Number         : %d\n", sense.sd_seg);
	printf("Flags                  :");
	if(sense.sd_sense & 0x80) printf(" Filemark");
	if(sense.sd_sense & 0x40) printf(" EOM");
	if(sense.sd_sense & 0x20) printf(" ILI");
	if(sense.sd_sense & 0x10) printf(" Reserved");
	if((sense.sd_sense & 0xf0) == 0) printf(" None");
	printf("\nSense Key              : %Xh (%s)\n",
		sense.sd_sense & SK_MASK, sense_keys[sense.sd_sense & SK_MASK]);
	if(sense.sd_err & 0x80) {
		p = (unsigned char *)&sense.sd_info;
		n = ((long)p[0] << 24) | ((long)p[1] << 16) | ((long)p[2] << 8) | (long)p[3];
		printf("Information            : %lxh or %ld\n", n, n);
	}

	if(sense.sd_asl < 11 - 7) return sk;
	p = (unsigned char *)&sense.sd_csinfo;
	n = ((long)p[0] << 24) | ((long)p[1] << 16) | ((long)p[2] << 8) | (long)p[3];
	printf("Command Specific Info  : %lxh or %ld\n",n, n);

	if(sense.sd_asl < 12 - 7) return sk;
	printf("Additional Sense Code  : %02Xh ", sense.sd_asc);
	if(sense.sd_asc >= 0x80) {
		printf("(Vendor Specific)\n");
	} else {
		for(s1 = scsi_asc; s1->desc; s1++) {
			if(s1->asc == sense.sd_asc) {
				break;
			}
		}
		if(!s1->desc) {
			printf("(Reserved)\n");
		} else if(sense.sd_asl < 13 - 7) {
			printf("(");
			printf(s1->desc, -1);
			printf(")\n");
		}
	}
	if(sense.sd_asl < 13 - 7) return sk;
	if(s1->ascq == -1) {
		printf("(");
		printf(s1->desc, sense.sd_ascq);
		printf(")\n");
	} else {
		if(sense.sd_ascq >= 0x80) {
			printf("(Vendor Specific)\n");
		} else {
			for(s2 = s1; s1->asc == s2->asc; s2++) {
				if(s2->ascq == sense.sd_ascq) {
					break;
				}
			}
			if(s1->asc == s2->asc) {
				printf("(%s)\n", s2->desc);
			} else {
				printf("(Reserved)\n");
			}
		}
	}
	printf("ASC Qualifier          : %02Xh\n", sense.sd_ascq);

	if(sense.sd_asl < 14 - 7) return sk;
	if(sense.sd_fruc) {
		printf("Field Replaceable Code : %02Xh\n", sense.sd_fruc);
	}
	if(sense.sd_asl < 15 - 7) return sk;
	if(sense.sd_sks_flags & 0x80) {
		printf("Sense key specific     : ");
		switch(sense.sd_sense & SK_MASK) {
		case SK_ILLEGAL:
			if(sense.sd_sks_flags & 0x40) {
				printf("Error in CDB (command descriptor block)\n");
			} else {
				printf("DATA (during DATA OUT PHASE)\n");
			}
			if(sense.sd_sks_flags & 0x30) {
				printf("Reserved SKS bits      :");
				if(sense.sd_sks_flags & 0x20) printf(" Reserved_bit_5");
				if(sense.sd_sks_flags & 0x10) printf(" Reserved_bit_4");
				printf("\n");
			}
			if(sense.sd_sks_flags & 0x08) {
				printf("Bit pointer is valid   : Bit %d, mask %02Xh\n",
				sense.sd_sks_flags & 0x07, 1 << (sense.sd_sks_flags & 0x07));
			} else if(sense.sd_sks_flags & 0x07) {
				printf("Bit pointer not valid  : %Xh\n", sense.sd_sks_flags & 0x07);
			}
			if(sense.sd_asl < 17 - 7) return sk;
			printf("Field pointer          : Error at byte offset %u\n",
				sense.sd_sks_data);
			break;
		case SK_RECOVERED:
		case SK_MEDIUM:
		case SK_HARDWARE:
			if(sense.sd_sks_flags & 0x7f) {
				printf("Reserved SKS fields    : %2Xh\n", sense.sd_sks_flags);
			}
			if(sense.sd_asl < 17 - 7) return sk;
			printf("Actual Retry Count     : %u\n", sense.sd_sks_data);
			break;
		case SK_NOT_RDY:
			if(sense.sd_sks_flags & 0x7f) {
				printf("Reserved SKS fields    : %02Xh\n", sense.sd_sks_flags);
			}
			if(sense.sd_asl < 17 - 7) return sk;
			printf("Progress Indication    : %u/65536 (%ld%)\n", sense.sd_sks_data,
				(long)sense.sd_sks_data / 65536L);
			break;
		default:
			printf("Reserved SKS fields    : %02Xh\n", sense.sd_sks_flags);
			if(sense.sd_asl < 17 - 7) return sk;
			printf("Reserved SKS info      : %u (%Xh)\n",
				sense.sd_sks_data, sense.sd_sks_data);
			break;
		}
	} else {
		if(sense.sd_asl < 17 - 7 || sense.sd_sks_flags || sense.sd_sks_data) {
			printf("Non-standard Sense key : %02Xh", sense.sd_sks_flags);
			if(sense.sd_asl >= 16 - 7) {
				p = (unsigned char *)&sense.sd_sks_data;
				printf(" %02Xh", p[0]);
				if(sense.sd_asl >= 17 - 7) {
					printf(" %02Xh", p[1]);
				}
			}
			printf("\n");
		}
	}
	if(sense.sd_asl < 18 - 7) return sk;
	printf("Additional Sense Bytes :");
	p = (unsigned char *)&sense.sd_asb;
	sense.sd_asl -= 18 - 7;
	while(sense.sd_asl--) {
		printf(" %02Xh", *p++);
	}
	printf("\n");
	return sk;
}

void dsp_stdinquiry(SCSI_INQUIRY *i) {
	char				*p;
	int					n;
	int					flag;
	struct scsi_types	*t;

	printf("Peripheral Qualifier   : %Xh ", (i->id_peripheral &
							(ID_QUAL_VENDOR|ID_QUAL_MASK)) >> 5);
	if(i->id_peripheral & ID_QUAL_VENDOR) {
		printf("(Vendor specific)\n");
	} else if((i->id_peripheral & ID_QUAL_MASK) == ID_QUAL_AVAIL) {
		printf("(Available)\n");
	} else if((i->id_peripheral & ID_QUAL_MASK) == ID_QUAL_NOTAVAIL) {
		printf("(Not Available)\n");
	} else if((i->id_peripheral & ID_QUAL_MASK) == ID_QUAL_RESERVED) {
		printf("(Reserved)\n");
	} else if((i->id_peripheral & ID_QUAL_MASK) == ID_QUAL_NOTSUP) {
		printf("(Not Supported)\n");
	} 
	printf("Peripheral Device Type : %02Xh ", i->id_peripheral & D_MASK);
	for(t = scsi_types; t->desc; t++) {
		if(t->peripheral == (i->id_peripheral & D_MASK)) {
			break;
		}
	}
	if(t->desc) {
		printf("(%s)\n", t->desc);
	} else {
		printf("(Reserved)\n");
	}
	printf("Removable              : %s", (i->id_rmb & ID_RMB) ? "Yes" : "No");
	if(i->id_rmb & ~ID_RMB) {
		printf(" Reserved=%02xh\n", i->id_rmb);
	} else {
		printf("\n");
	}
	if(i->id_ver & 0xc0) {
		printf("ISO version            : %d", i->id_ver >> 6);
	}
	if(i->id_ver & 0x38) {
		printf("ECMA version           : %d", (i->id_ver & 0x38) >> 3);
	}
	printf("ANSI version           : ");
	switch(i->id_ver & 0x07) {
	case 0:
		printf("Might/might not be ANSI conforming\n");
		break;
	case 1:
		printf("X3.131-1986 (SCSI1)\n");
		break;
	case 2:
		printf("Draft 10 (SCSI2)\n");
		break;
	default:
		printf("Unknown ANSI version %d\n", i->id_ver & 0x07);
		break;
	}
	printf("Response data version  : ");
	switch(i->id_rdf & 0x0f) {
	case 0:
		printf("SCSI-1\n");
		break;
	case 1:
		printf("CCS\n");
		break;
	case 2:
		printf("SCSI-2\n");
		break;
	default:
		printf("Reserved (%d)\n", i->id_rdf & 0x0f);
		break;
	}
	if(i->id_rdf & 0xf0) {
		printf("Standard flags (Scsi-2):");
		if(i->id_rdf & 0x80) printf(" AENC");
		if(i->id_rdf & 0x40) printf(" TrmIOP");
		if(i->id_rdf & 0x20) printf(" Reserved-BIT5");
		if(i->id_rdf & 0x10) printf(" Reserved-BIT4");
		printf("\n");
	}
	if(i->id_adlen < 5 - 4) return;
	if(i->id_res1) {
		printf("Reserved byte 5        : %02Xh\n", i->id_res1);
	}
	if(i->id_adlen < 6 - 4) return;
	if(i->id_res2) {
		printf("Reserved byte 6        : %02Xh\n", i->id_res2);
	}
	if(i->id_adlen < 7 - 4) return;
	printf("Flags                  :");
	if(i->id_flags) {
		if(i->id_flags & 0x80) printf(" RelAdr");
		if(i->id_flags & 0x40) printf(" WBus32");
		if(i->id_flags & 0x20) printf(" WBus16");
		if(i->id_flags & 0x10) printf(" Sync");
		if(i->id_flags & 0x08) printf(" Linked");
		if(i->id_flags & 0x04) printf(" Reserved");
		if(i->id_flags & 0x02) printf(" CmdQue");
		if(i->id_flags & 0x01) printf(" SftRe");
		printf("\n");
	} else {
		printf(" None\n");
	}
	if(i->id_adlen < 15 - 4) return;
	printf("Vendor Identification  : \"%-8.8s\"\n", i->id_vend_id);
	if(i->id_adlen < 31 - 4) return;
	printf("Product Identification : \"%-16.16s\"\n", i->id_prod_id);
	if(i->id_adlen < 35 - 4) return;
	printf("Product Revision Level : \"%-4.4s\"\n", i->id_prod_rev);
	if(i->id_adlen < 36 - 4) return;
	printf("Vendor Specific bytes  :");
	p = (unsigned char *)i->id_vend_spc;
	n = min(20, i->id_adlen - 36 + 4);
	flag = 0;
	while(n-- > 0 && flag < 2) {
		if((flag && *p) || (flag == 0 && (*p <= ' ' || *p >= 0x7f))) {
			flag = *p ? 2 : 1;
		}
		p++;
	}
	p = (unsigned char *)i->id_vend_spc;
	n = min(20, i->id_adlen - 36 + 4);
	if(flag < 2) printf(" \"");
	while(n-- > 0) {
		if(flag == 2) {
			printf(" %02Xh", *p);
			if(*p > ' ' && *p < 0x7f) {
				printf("(%c)", *p);
			}
		} else {
			printf("%c", *p);
		}
		p++;
	}
	if(flag < 2) printf("\"");
	printf("\n");
	if(i->id_adlen < 56 - 4) return;
	p = (unsigned char *)i->id_res3;
	n = i->id_adlen - 56 + 4;
	while(n-- > 0) {
		if(*p++) break;
	}
	if(n < 0) return;
	printf("Reserved bytes         :");
	p = (unsigned char *)i->id_res3;
	n = i->id_adlen - 56 + 4;
	while(n-- > 0) {
		printf(" %02Xh", *p++);
	}
	printf("\n");
}

dsp_inquiry00(void *information) {
	char				*p = information;
	int					i;

	printf("Supported inquiry pages:");
	if(p[3] == 0) {
		printf(" None\n");
		return;
	}
	for(i = 3; i <= p[3] + 3; i++) {
		printf(" 0x%02x", p[i]);
	}
	printf("\n");
}

void cmd_inquiry(void) {
	char				*p;
	int					i;
	int					page = -1;

    scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
	if(p = strtok(0, " \n\t")) {
		page = strtol(p, 0, 0);
		if(page < 0 || page > 0xff) {
			printf("Page number (%02Xh) out of range (00h-FFh)\n");
			return;
		}
	}
	scsicdb[0] = SCSI_OP(0, SC_INQUIRY);
	scsicdb[4] = min(0xFF, sizeof scsibuf);
	if(page != -1) {
		scsicdb[1] |= 0x01;
		scsicdb[2] = page;
	}
	scsicmd.uscsi_cdblen = 6;
	if(scsi_cmd(1, "inquiry") == -1) {
		return;
	}
	switch(page) {
	case -1:		// Standard Inquiry
		dsp_stdinquiry((SCSI_INQUIRY *)scsibuf);
		break;
	case 0x00:		// Supported Vital Products Data Pages
		dsp_inquiry00(scsibuf);
		break;
	default:
		printf("Vital Products Page    : %Xh\n");
		printf("Vital Products bytes   :");
		for(i = 0; i <  p[3] + 3; i++) {
			printf(" 0x%02x", p[i]);
		}
		printf("\n");
		break;
	}
}

void cmd_info(void) {
	struct scsi_types		*t;
	struct uscsi_info		info;

	if(qnx_ioctl(fd, USCSIINFO, 0, 0, &info, sizeof info) == -1) {
		printf("Couldn't get information\n");
		return;
	}
	printf("Adaptor path ID        : %d\n", info.path);
	printf("Device SCSI target ID  : %d\n", info.target);
	printf("Device SCSI LUN        : %d\n", info.lun);
	printf("Peripheral Device Type : %02Xh ", info.peripheral);
	for(t = scsi_types; t->desc; t++) {
		if(t->peripheral == info.peripheral) {
			break;
		}
	}
	if(t->desc) {
		printf("(%s)\n", t->desc);
	} else {
		printf("(Reserved)\n");
	}
}

void cmd_defect(void) {
	unsigned char		*p;
	unsigned			size = min(0xFFFF, sizeof scsibuf);
	unsigned			num;
	unsigned			wrap;
	int					missing;

    scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
	scsicdb[0] = SC1_RD_DEF;
	scsicdb[2] = 0x05;
	scsicdb[7] = size >> 8;
	scsicdb[8] = size & 0xff;
	if(p = strtok(0, " \n\t")) {
		if(strchr(p, 'g') || strchr(p, 'G')) {
			scsicdb[2] |= 0x08;
		}
		if(strchr(p, 'p') || strchr(p, 'P')) {
			scsicdb[2] |= 0x10;
		}
	} else {
		scsicdb[2] |= 0x18;
	}
	scsicmd.uscsi_cdblen = 10;
	if(scsi_cmd(1, "read defect data") == -1) {
		return;
	}
	printf("Defect list format     : %Xh\n", scsibuf[1] & 0x07);
	printf("Defect lists           :");
	if(scsibuf[1] & 0x10) printf(" Primary");
	if(scsibuf[1] & 0x08) printf(" Grown");
	if((scsibuf[1] & 0x18) == 0) printf(" Header only");
	printf("\n");
	missing = (scsibuf[2] << 8) | scsibuf[3];
	size = min(size - 4, missing);
	missing -= size;
	switch(scsibuf[1] & 0x07) {
	case 0x00:
		printf("Bad Blocks             :" );
		size /= 4;
		missing /= 4;
		wrap = 3;
		break;
	case 0x04:
		printf("Bad track,head,offset  :" );
		size /= 8;
		missing /= 8;
		wrap = 2;
		break;
	case 0x05:
		printf("Bad track,head,sector  :" );
		size /= 8;
		missing /= 8;
		wrap = 3;
		break;
	default:
		printf(" Unable to display format\n");
		return;
	}
	num = wrap;
	p = &scsibuf[4];
	if(size == 0) {
		printf(" None");
	}
	while(size--) {
		unsigned long		block;
		unsigned long		cyl;
		unsigned			head;

		if(num == -1) {
			printf("                        ");
			num = wrap;
		}
		switch(scsibuf[1] & 0x07) {
		case 0x00:
			block = ((long)p[0] << 24l) | ((long)p[1] << 16l) |
					((long)p[2] << 8l) | (long)p[3];
			p += 4;
			if(block == 0) continue;
			printf(" %08lXh" );
			break;
		case 0x04:
			cyl = ((long)p[0] << 16) | ((long)p[1] << 8) |
					(long)p[2];
			head = p[3];
			block = ((long)p[4] << 24) | ((long)p[5] << 16) |
					((long)p[6] << 8) | (long)p[7];
			p += 8;
			if(cyl == 0 && head == 0 && block == 0) continue;
			printf(" %4.4lu,%02u,%-9lu", cyl, head, block);
			break;
		case 0x05:
			cyl = ((long)p[0] << 16) | ((long)p[1] << 8) |
					(long)p[2];
			head = p[3];
			block = ((long)p[4] << 24) | ((long)p[5] << 16) |
					((long)p[6] << 8) | (long)p[7];
			p += 8;
			if(cyl == 0 && head == 0 && block == 0) continue;
			printf(" %4lu,%02u,%-3lu", cyl, head, block);
			break;
		}
		if(num-- == 0) {
			printf("\n");
		}
	}
	if(num != -1) printf("\n");
	if(missing > 0) {
		printf("Number of missing items: %d\n", missing);
	}
}

dsp_geometry(char *p) {
	long				n;
	long				cyls;

	cyls = ((long)p[2] << 16) | ((long)p[3] << 8) | (long)p[4];
	printf("Number of cylinders      %ld\n", cyls);
	if(p[1] < 5 - 1) return;
	printf("Number of heads          %d\n", p[5]);
	if(p[1] < 8 - 1) return;
	n = ((long)p[6] << 16) | ((long)p[7] << 8) | (long)p[8];
	printf("Write precomp cylinder   %ld", n);
	if(n >= cyls) {
		printf(" (disabled)\n");
	} else {
		printf("\n");
	}
	if(p[1] < 11 - 1) return;
	n = ((long)p[9] << 16) | ((long)p[10] << 8) | (long)p[11];
	printf("Write low current cyl    %ld", n);
	if(n >= cyls) {
		printf(" (disabled)\n");
	} else {
		printf("\n");
	}
	if(p[1] < 13 - 1) return;
	n = ((long)p[12] << 8) | (long)p[13];
	printf("Drive step rate          %ld", n);
	if(n == 0) {
		printf(" (default)\n");
	} else {
		n *= 100;
		printf(" (%ld ns)\n", n);
	}
	if(p[1] < 16 - 1) return;
	n = ((long)p[14] << 16) | ((long)p[15] << 8) | (long)p[16];
	if(p[14] & 0x80) n |= 0xff000000;
	printf("Landing zone cylinder    %ld", n);
	if(n) {
		printf("\n");
	} else {
		printf(" (default)\n");
	}
	if(p[1] < 17 - 1) return;
	printf("Rotational posn locking  ");
	switch(p[17] & 0x03) {
	case 0:
		printf("Disabled or not supported\n");
		break;
	case 1:
		printf("spindle slave\n");
		break;
	case 2:
		printf("spindle master\n");
		break;
	case 3:
		printf("spindle master control\n");
		break;
	}
	if(p[1] < 18 - 1) return;
	if(p[18]) {
		printf("Rotational offset        %d/256 revolution\n", p[18]);
	}
	if(p[1] < 21 - 1) return;
	n = ((long)p[20] << 8) | (long)p[21];
	printf("Rotational speed         %d rpm\n", n);
}

void cmd_geometry(void) {
	char				*p;

    scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
	scsicdb[0] = SC_MSENSE;
	scsicdb[2] = 0x04;
	scsicmd.uscsi_buflen = scsicdb[4] = (unsigned char)max(12, min(0xff, sizeof scsibuf));
	scsicmd.uscsi_cdblen = 6;
	if(scsi_cmd(1, "read drive geometry") == -1) {
		return;
	}
	if(scsibuf[0] < 4 || !(p = scsibuf + (scsibuf[3] + 4)) ||
		(p[0] & 0x3f) != 0x04 || p[1] < 4 - 1) {
		printf("Couldn't understand geometry page\n");
		return;
	}
	dsp_geometry(p);
}

void cmd_medium(void) {
	char				*p;
	int					medium;

	scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
	scsicdb[0] = SC_MSENSE;
	scsicdb[4] = scsicmd.uscsi_buflen = 12;
	scsicmd.uscsi_cdblen = 6;
	if(scsi_cmd(1, "read medium") == -1) {
		return;
	}
	printf("Medium code            :");
	if(scsibuf[0] < 2) {
		printf(" Missing\n");
		return;
	} else {
		printf(" %0Xh\n", scsibuf[1]);
	}
	if(p = strtok(0, "=,\n\t")) {
		medium = strtol(p, 0, 16);
		printf("Changing to medium     : %Xh\n", medium);
		scsicmd.uscsi_flags = 0;
		scsibuf[1] = medium;
		scsicdb[0] = SC_MSELECT;
		scsicdb[4] = scsicmd.uscsi_buflen = scsibuf[0] + 1;
		scsibuf[0] = 0;
		scsicmd.uscsi_cdblen = 6;
		if(scsi_cmd(1, "set medium") == -1) {
			return;
		}
	}
}

void cmd_density(void) {
	char				*p;
	int					i;
	int					density;

    scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
	scsicdb[0] = SC_MSENSE;
	scsicdb[4] = scsicmd.uscsi_buflen = 12;
	scsicmd.uscsi_cdblen = 6;
	if(scsi_cmd(1, "read density") == -1) {
		return;
	}
	printf("Density code           :");
	if(scsibuf[0] < 5 || scsibuf[3] < 1) {
		printf(" Missing\n");
		return;
	} else {
		for(i = 0; i < scsibuf[3]; i += 8) {
			printf(" %Xh", scsibuf[i + 4]);
		}
		printf("\n");
	}
	if(p = strtok(0, "=,\n\t")) {
		density = strtol(p, 0, 16);
		printf("Changing to code       : %Xh\n", density);
		for(i = 0; i < scsibuf[3]; i += 8) {
			scsibuf[i + 4] = density;
		}
		scsicmd.uscsi_flags = 0;
		scsicdb[0] = SC_MSELECT;
		scsicdb[4] = scsicmd.uscsi_buflen = scsibuf[0] + 1;
		scsibuf[0] = 0;
		scsicmd.uscsi_cdblen = 6;
		if(scsi_cmd(1, "set density") == -1) {
			return;
		}
	}
}

void cmd_sense(void) {
	unsigned char		*p;
	int					i;
	int					length1, length2;
	unsigned char		medium, device;
	struct uscsi_info	info;
	unsigned			peripheral;
	unsigned			size = min(0xFF, sizeof scsibuf);

	if(p = strtok(0, " \n\t")) {
	}
	if(qnx_ioctl(fd, USCSIINFO, 0, 0, &info, sizeof info) == -1) {
		printf("Couldn't device type information\n");
		return;
	}
	peripheral = info.peripheral;
    scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
	scsicdb[0] = SC_MSENSE;
	scsicdb[1] = 0x00; // 0x08 to disable block descriptors
	scsicdb[2] = 0x00 | 0x3f; // Current values, all pages
	scsicmd.uscsi_buflen = scsicdb[4] = size;
	scsicmd.uscsi_cdblen = 6;
	if(scsi_cmd(0, "read sense pages (6)") == -1) {
		size = min(0xFFFF, sizeof scsibuf);
	    scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
		scsicdb[0] = SC2_MSENSE;
		scsicdb[1] = 0x00; // 0x08 to disable block descriptors
		scsicdb[2] = 0x00 | 0x3f; // Current values, all pages
		scsicdb[7] = size >> 8;
		scsicdb[8] = size & 0xff;
		scsicmd.uscsi_buflen = size;
		scsicmd.uscsi_cdblen = 10;
		if(scsi_cmd(0, "read sense pages (10)") == -1) {
			size = min(0xFF, sizeof scsibuf);
 			scsicmd.uscsi_flags = USCSI_READ|USCSI_NOFLUSH;
			scsicdb[0] = SC_MSENSE;
			scsicdb[1] = 0x00; // 0x08 to disable block descriptors
			scsicdb[2] = 0x00;
			scsicmd.uscsi_buflen = scsicdb[4] = size;
			scsicmd.uscsi_cdblen = 6;
			if(scsi_cmd(1, "read sense pages (SCSI-1)") == -1) {
				return;
			} else {
				length1 = scsibuf[3];
				length2 = scsibuf[0] - 3 - length1;
				medium = scsibuf[1];
				device = scsibuf[2];
				p = &scsibuf[4];
			}
		} else {
			length1 = (scsibuf[6] << 8) | scsibuf[7];
			length2 = ((scsibuf[0] << 8) | scsibuf[1]) - 6 - length1;
			medium = scsibuf[2];
			device = scsibuf[3];
			p = &scsibuf[8];
		}
	} else {
		length1 = scsibuf[3];
		length2 = scsibuf[0] - 3 - length1;
		medium = scsibuf[1];
		device = scsibuf[2];
		p = &scsibuf[4];
	}
	switch(peripheral) {
	case D_DIR_ACC:
		printf("Medium type            : %Xh ", medium);
		switch(medium) {
		case 0x00:
			printf("Default medium type (currently mounted)\n");
			break;
		default:
			if(medium >= 0x80) {
				printf("Vendor-Specific\n");
			} else {
				printf("Reserved\n");
			}
		}
		printf("Device-specific parm   : %Xh", device);
		if(device & 0x80) {
			printf(" (Write Protect)");
		}
		if(device & 0x10) {
			printf(" (Cache)");
		}
		printf("\n");
		break;
	case D_WORM:
	case D_OPTICAL:
		printf("Medium type            : %Xh ", medium);
		switch(medium) {
		case 0x00:
			printf("Default (only one medium type supported)\n");
			break;
		case 0x01:
			printf("Optical Read Only medium\n");
			break;
		case 0x02:
			printf("Optical Write Once medium\n");
			break;
		case 0x03:
			printf("Optical Reversible or Erasable medium\n");
			break;
		case 0x04:
			printf("Combination of Read Only and Write Once medium\n");
			break;
		case 0x05:
			printf("Combination of Read Only and Reversible or Eraseable medium\n");
			break;
		case 0x06:
			printf("Combination of Write Once and Reversible or Eraseable medium\n");
			break;
		default:
			if(medium >= 0x80) {
				printf("Vendor-Specific\n");
			} else {
				printf("Reserved\n");
			}
		}
		printf("Device-specific parm   : %Xh", device);
		if(device & 0x80) {
			printf(" (Write Protect)");
		}
		if(device & 0x10) {
			printf(" (Cache)");
		}
		if(device & 0x01) {
			printf(" (Enable blank check)");
		}
		printf("\n");
		break;
	case D_READ_ONLY:
		printf("Medium type            : %Xh ", medium);
		switch(medium) {
		case 0x00:
			printf("Default (only one type supported)\n");
			break;
		case 0x01:
			printf("120mm CD-ROM data only\n");
			break;
		case 0x02:
			printf("120mm CD-DA data only\n");
			break;
		case 0x03:
			printf("120mm CD-ROM data and audio combined\n");
			break;
		case 0x05:
			printf("80mm CD-ROM data only\n");
			break;
		case 0x06:
			printf("80mm CD-DA data only\n");
			break;
		case 0x07:
			printf("80mm CD-ROM data and audio combined\n");
			break;
		default:
			if(medium >= 0x80) {
				printf("Vendor unique\n");
			} else {
				printf("Reserved\n");
			}
		}
		printf("Device-specific parm   : %Xh", device);
		if(device & 0x10) {
			printf(" (Cache)");
		}
		if(device & 0x01) {
			printf(" (Enable blank check)");
		}
		printf("\n");
		break;
	case D_SEQ_ACC:
		if(medium) {
			printf("Medium type            : %Xh (reserved)\n", medium);
		}
		printf("Device-specific parm   : %Xh", device);
		if(device & 0x80) {
			printf(" (Write Protect)");
		}
		printf("\nBuffered mode          : %Xh\n", (device >> 4) & 3);
		printf("Speed                  : %Xh\n", device & 0x0f);
		break;
	default:
		printf("Medium type            : %Xh\n", medium);
		printf("Device-specific parm   : %Xh\n", device);
		break;
	}
	while(length1 > 0) {
		if(length1 < 8) {
			printf("Incomplete block desc  :");
			while(length1--) {
				printf(" %Xh", *p++);
			}
			printf("\n");
			break;
		} else {
			switch(peripheral) {
			case D_DIR_ACC:
				if(p[0]) {
					printf("Density Code           : %Xh (reserved)\n", p[0]);
				}
				break;
			case D_SEQ_ACC:
				printf("Density Code           : %Xh ", p[0]);
				switch(p[0]) {
				case 0x00:
					printf("Default density\n");
					break;
				case 0x7f:
					printf("No change from previous\n");
					break;
				default:
					if(p[0] < 0x80) {
						printf("Reserved\n");
					} else {
						printf("Vendor unique\n");
					}
					break;
				}
				break;
			case D_WORM:
			case D_OPTICAL:
				printf("Density Code           : %Xh ", p[0]);
				switch(p[0]) {
				case 0x00:
					printf("Default density (currently mounted)\n");
					break;
				default:
					if(p[0] < 0x80) {
						printf("Reserved\n");
					} else {
						printf("Vendor unique\n");
					}
					break;
				}
				break;
			case D_READ_ONLY:
				printf("Density Code           : %Xh ", p[0]);
				switch(p[0]) {
				case 0x00:
					printf("Default Density Code\n");
					break;
				case 0x01:
					printf("User Data Only - 2048 bytes/sec\n");
					break;
				case 0x02:
					printf("User Data plus auxiliary data field - 2336 bytes/sec\n");
					break;
				case 0x03:
					printf("4 byte tag, user Data plus auxiliary data field - 2340 bytes/sec\n");
					break;
				case 0x04:
					printf("Audio Information Only - 1/75 second/block\n");
					break;
				default:
					if(p[0] < 0x80) {
						printf("Reserved\n");
					} else {
						printf("Vendor unique\n");
					}
					break;
				}
				break;
			default:
				printf("Density Code           : %Xh\n", p[0]);
				break;
			}
			printf("Number of blocks       : %ld\n", ((long)p[1] << 16) | ((long)p[2] << 8) | (long)p[3]);
			if(p[4]) {
				printf("Reserved               : %Xh\n", p[4]);
			}
			printf("Block length           : %ld\n", ((long)p[5] << 16) | ((long)p[6] << 8) | (long)p[7]);
			p += 8;
			length1 -= 8;
		}
	}
	while(length2 > 0) {
		int					size;
		int					page;

		if(length2 < 2) {
			printf("Partial code page      :");
			while(length2--) {
				printf(" %Xh", *p++);
			}
			printf("\n");
			break;
		}
		page = p[0] & 0x3f;
		size = p[1];
		printf("Page Code              : %Xh", page);
		if(page == 0) {
			size = length2 - 2;
		} else {
			if(p[0] & 0x80) {
				printf(" (Saveable)");
			}
			if(p[0] & 0x40) {
				printf(" (Reserved)");
			}
		}
		length2 -= 2;
		if(length2 < size) {
			printf("\nPartial code page      :");
			while(length2--) {
				printf(" %Xh", *p++);
			}
			printf("\n");
			break;
		}
		switch(page) {
		case 0x00:
			printf(" Vendor Specific Page\n");
			printf("Vendor specific bytes   ");
			for(i = 0; i < size + 2; i++) {
				printf(" %Xh", p[i]);
			}
			printf("\n");
			break;
		case 0x01:
			switch(peripheral) {
			case D_DIR_ACC:
			case D_WORM:
			case D_OPTICAL:
				printf(" Read-Write Error Recovery Page\n");
				printf("Flags                   ");
				if(p[2] == 0) {
					printf(" None");
				}
				if(p[2] & 0x80) {
					printf(" (Automatic Write Reallocation Enabled)");
				}
				if(p[2] & 0x40) {
					printf(" (Automatic Read Reallocation Enabled)");
				}
				if(p[2] & 0x20) {
					printf(" (Transfer Block)");
				}
				if(p[2] & 0x10) {
					printf(" (Read Continuous)");
				}
				if(p[2] & 0x08) {
					printf(" (Enable Early Recovery)");
				}
				if(p[2] & 0x04) {
					printf(" (Post Error)");
				}
				if(p[2] & 0x02) {
					printf(" (Disable Transfer on Error)");
				}
				if(p[2] & 0x01) {
					printf(" (Disable Correction)");
				}
				printf("\n");
				printf("Read Retry Count         %d\n", p[3]);
				if(p[4]) {
					printf("Correction Span          %d bits\n", p[4]);
				}
				if(p[5]) {
					printf("Head Offset Count        %d\n", (signed char)p[5]);
				}
				if(p[6]) {
					printf("Data Strobe Offset Count %d\n", (signed char)p[6]);
				}
				if(p[7]) {
					printf("Reserved                 %d\n", p[7]);
				}
				if(p[8]) {
					printf("Write Retry Count        %d\n", p[8]);
				}
				if(p[9]) {
					printf("Reserved                 %d\n", p[9]);
				}
				if(p[10] || p[11]) {
					printf("Recovery Time Limit      %ums\n", (p[10] << 8) | p[11]);
				}
				break;
			case D_SEQ_ACC:
				printf(" Read-Write Error Recovery Page\n");
				goto display;
			case D_READ_ONLY:
				printf(" Read Error Recovery Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x02:
			printf(" Disconnect-Reconnect Page\n");
			printf("Buffer Full Ratio        ");
			if(p[2]) {
				printf("%d/256 (%d%%)\n", p[2], (p[2] * 100) / 256);
			} else {
				printf("Target determined\n");
			}
			printf("Buffer Empty Ratio       ");
			if(p[3]) {
				printf("%d/256 (%d%%)\n", p[3], (p[3] * 100) / 256);
			} else {
				printf("Target determined\n");
			}
			printf("Bus Inactivity Limit     ");
			if(p[4] || p[5]) {
				printf("%ldus\n", (long)((p[4] << 8) | p[5]) * 100);
			} else {
				printf("No Limit\n");
			}
			printf("Disconnect Time Limit    ");
			if(p[6] || p[7]) {
				printf("%ldus\n", (long)((p[6] << 8) | p[7]) * 100);
			} else {
				printf("No Limit\n");
			}
			printf("Connect Time Limit       ");
			if(p[8] || p[9]) {
				printf("%ldus\n", (long)((p[8] << 8) | p[9]) * 100);
			} else {
				printf("No Limit\n");
			}
			printf("Maximum Burst Size       ");
			if(p[10] || p[11]) {
				printf("%u bytes\n", (p[10] << 8) | p[11]);
			} else {
				printf("No Limit\n");
			}
			printf("Data Xfer Discon Cntrl   %d ", p[12] & 0x03);
			switch(p[12] & 0x03) {
			case 0:
				printf("Disconnect Control not used\n");
				break;
			case 1:
				printf("No Disconnect while data transfer\n");
				break;
			case 2:
				printf("Reserved\n");
				break;
			case 3:
				printf("No Disconnect until command complete\n");
				break;
			}
			if((p[12] & ~0x03) || p[13] || p[14] || p[15]) {
				printf("Reserved bytes           %Xh %Xh %Xh %Xh\n", p[12], p[13], p[14], p[15]);
			}
			break;

		case 0x03:
			switch(peripheral) {
			case D_DIR_ACC:
				printf(" Format Device Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x04:
			switch(peripheral) {
			case D_DIR_ACC:
				printf(" Rigid Disk Drive Geometry Page\n");
				dsp_geometry(p);
				break;
			default:
				goto unknown;
			}
			break;

		case 0x05:
			switch(peripheral) {
			case D_DIR_ACC:
				printf(" Flexible Disk Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x06:
			switch(peripheral) {
			case D_WORM:
			case D_OPTICAL:
				printf(" Optical Memory Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x07:
			switch(peripheral) {
			case D_DIR_ACC:
			case D_WORM:
			case D_OPTICAL:
				printf(" Verify Error Recovery Page\n");
				goto display;
			case D_READ_ONLY:
				printf(" Verify Error Recovery Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x08:
			switch(peripheral) {
			case D_DIR_ACC:
			case D_WORM:
			case D_OPTICAL:
			case D_READ_ONLY:
				printf(" Caching Page\n");
				if(p[2] & 0xf8) {
					printf("Reserved                 %Xh\n", p[2]);
				}
				printf("Flags                   ");
				if(p[2] & 0x07) {
					if(p[2] & 0x04) {
						printf(" (Write Cache Enable)");
					}
					if(p[2] & 0x02) {
						printf(" (Multipy Factor)");
					}
					if(p[2] & 0x01) {
						printf(" (Read Cache Disable)");
					}
				} else {
					printf(" None");
				}
				printf("\n");
				printf("Demand Write Retent Pri  ");
				switch(p[3] & 0x0f) {
				case 0x0:
					printf("Not distinguished\n");
					break;
				case 0x1:
					printf("Read has priority\n");
					break;
				case 0xf:
					printf("Read Cache locked\n");
					break;
				default:
					printf("%Xh Reserved\n", p[3] & 0x0f);
					break;
				}
				printf("Demand Read Retent Pri   ");
				switch(p[3] >> 4) {
				case 0x0:
					printf("Not distinguished\n");
					break;
				case 0x1:
					printf("Write has priority\n");
					break;
				case 0xf:
					printf("Write Cache locked\n");
					break;
				default:
					printf("%Xh Reserved\n", p[3] >> 4);
					break;
				}
				printf("Disable Pre-fetch len    ");
				if(p[4] || p[5]) {
					printf("%d ", (p[4] << 8) | p[5]);
					if(p[2] & 0x02) {
						printf("times xfer len\n");
					} else {
						printf("blocks\n");
					}
				} else {
					printf("No pre-fetch\n");
				}
				printf("Minimum Pre-fetch        ");
				if(p[6] || p[7]) {
					printf("%d ", (p[6] << 8) | p[7]);
					if(p[2] & 0x02) {
						printf("times xfer len\n");
					} else {
						printf("blocks\n");
					}
				} else {
					printf("Stop pre-fetch when other command issued\n");
				}
				printf("Maximum Pre-fetch        ");
				printf("%d ", (p[8] << 8) | p[9]);
				if(p[2] & 0x02) {
					printf("times xfer len\n");
				} else {
					printf("blocks\n");
				}
				printf("Maximum Pre-fetch Ceiling");
				printf("%d ", (p[10] << 8) | p[11]);
				if(p[2] & 0x02) {
					printf("times xfer len\n");
				} else {
					printf("blocks\n");
				}
				break;

			default:
				goto unknown;
			}
			break;

		case 0x09:
			printf(" Peripheral Device Page\n");
			printf("Interface Identifier     %Xh\n", (p[2] << 8) | p[3]);
			if(p[4] || p[5] || p[6] || p[7]) {
				printf("Reserved fields          %Xh %Xh %Xh %Xh\n", p[4], p[5], p[6], p[7]);
			}
			if(8 - 2 < size) {
				printf("Vendor Specific         ");
				for(i = 8; i < size + 2; i++) {
					printf(" %Xh", p[i + 2]);
				}
				printf("\n");
			}
			break;

		case 0x0a:
			printf(" Control Mode Page\n");
			printf("Flags                    %Xh %Xh %Xh", p[2], p[3], p[4]);
			if(p[2] & 0x01) {
				printf(" (Report log excption)");
			}
			if(p[3] & 0x02) {
				printf(" (Queue Error)");
			}
			if(p[3] & 0x01) {
				printf(" (Disable queueing)");
			}
			if(p[4] & 0x80) {
				printf(" (Enable Extended Contingent Allegeince)");
			}
			if(p[4] & 0x04) {
				printf(" (Ready AEN Permission)");
			}
			if(p[4] & 0x02) {
				printf(" (Unit Attention AEN permission)");
			}
			if(p[4] & 0x01) {
				printf(" (Error AEN permission)");
			}
			printf("\n");
			printf("Queue Algorithm Modifier %Xh ", p[3] >> 4);
			switch(p[3] >> 4) {
			case 0x0:
				printf("Guaranteed data integrity\n");
				break;
			case 0x1:
				printf("Unrestricted re-ording allowed\n");
				break;
			default:
				if((p[3] >> 4) < 0x8) {
					printf("Reserved\n");
				} else {
					printf("Vendor Specific\n");
				}
				break;
			}
			printf("Ready AEN Holdoff period %Xh\n", (p[6] << 8) | p[7]);
			break;

		case 0x0b:
			switch(peripheral) {
			case D_DIR_ACC:
			case D_WORM:
			case D_OPTICAL:
			case D_READ_ONLY:
				printf(" Medium Types Supported Page\n");
				if(p[2] || p[3]) {
					printf("Reserved fields          %Xh %Xh\n", p[2], p[3]);
				}
				printf("Medium type 1 supported  %Xh\n", p[4]);
				printf("Medium type 2 supported  %Xh\n", p[5]);
				printf("Medium type 3 supported  %Xh\n", p[6]);
				printf("Medium type 4 supported  %Xh\n", p[7]);
				break;
			default:
				goto unknown;
			}
			break;

		case 0x0c:
			switch(peripheral) {
			case D_DIR_ACC:
				printf(" Notch Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x0d:
			switch(peripheral) {
			case D_READ_ONLY:
				printf(" CD-ROM Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x0e:
			switch(peripheral) {
			case D_READ_ONLY:
				printf(" CD-ROM Audio Control Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x10:
			switch(peripheral) {
			case D_SEQ_ACC:
				printf(" Device Configuration Page\n");
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x11:
			switch(peripheral) {
			case D_SEQ_ACC:
				printf(" 1st Medium Partition Page\n");
medium_partition:
				goto display;
			default:
				goto unknown;
			}
			break;

		case 0x12:
			switch(peripheral) {
			case D_SEQ_ACC:
				printf(" 2nd Medium Partition Page\n");
				goto medium_partition;
			default:
				goto unknown;
			}
			break;

		case 0x13:
			switch(peripheral) {
			case D_SEQ_ACC:
				printf(" 3rd Medium Partition Page\n");
				goto medium_partition;
			default:
				goto unknown;
			}
			break;

		case 0x14:
			switch(peripheral) {
			case D_SEQ_ACC:
				printf(" 4th Medium Partition Page\n");
				goto medium_partition;
			default:
				goto unknown;
			}
			break;

		default:
unknown:
			if(page >= 0x20 || page == 0x00) {
				printf(" Vendor Specific Page\n");
			} else {
				printf(" Unknown Code Page\n");
			}
display:
			printf("Code page bytes         ");
			for(i = 0; i < size; i++) {
				printf(" %Xh", p[i + 2]);
			}
			printf("\n");
			break;
		}
		length2 -= size; 
		p += size + 2;
	}
}

void cmd_scan(void) {
	int						i;
	struct dirent			*dp;
    DIR						*dirp;
	struct scsi_types		*t;
	struct uscsi_info		info;
	char					dev[PATH_MAX+1];

	if(fd != -1) {
		close(fd);
		fd = -1;
	}
	sprintf(path, "//%ld/dev", nid ? nid : getnid());
	if(!(dirp = opendir(path))) {
		fprintf(stderr, "Error opening %s (%s)\n", path, strerror(errno));
		return;
	}
	printf("Device            PathID Target Lun   Type\n");
	i = strlen(path);
	path[i++] = '/';
	while(dp = readdir(dirp)) {
		strcpy(&path[i], dp->d_name);
  	
		if(!(dp->d_stat.st_status & _FILE_USED)) {
			if(stat(path, &dp->d_stat) == -1) {
				printf("%-20s Could not stat() (%s)\n", path, strerror(errno));
				continue;
			}
		}
		if(!S_ISBLK(dp->d_stat.st_mode)) {
			continue;
		}
		if(fsys_get_mount_dev(path, dev) != -1 || errno != EINVAL) {
			continue;
		}
		if((fd = open(path, O_RDONLY)) == -1) { 
			printf("%-20s Could not open for read (%s)\n", path, strerror(errno));
			continue;
		}
		if(qnx_ioctl(fd, USCSIINFO, 0, 0, &info, sizeof info) == -1) {
			close(fd);
			continue;
		}
		for(t = scsi_types; t->desc; t++) {
			if(t->peripheral == info.peripheral) {
				break;
			}
		}
		printf("%-20s %2d    %d     %d    %s (%02X)\n",
			path, info.path, info.target, info.lun,
			t->desc ? t->desc : "(Reserved)", info.peripheral);
		close(fd);
	}
	closedir(dirp);
	fd = -1;
}

struct cmds {
	char		*name;
	int			oflag;
	void		(*func)(void);
} cmds[] = {
	{	"scan",			-1,			cmd_scan		},
	{	"lock",			O_RDONLY,	cmd_lock		},
	{	"unlock",		O_RDONLY,	cmd_unlock		},
	{	"start",		O_RDONLY,	cmd_start		},
	{	"stop",			O_RDONLY,	cmd_stop		},
	{	"load",			O_RDONLY,	cmd_load		},
	{	"unload",		O_RDONLY,	cmd_unload		},
	{	"eject",		O_RDONLY,	cmd_unload		},
	{	"defect", 		O_RDONLY,	cmd_defect		},
	{	"density", 		O_RDONLY,	cmd_density		},
	{	"medium", 		O_RDONLY,	cmd_medium		},
//	{	"diagnostic",	O_RDONLY,	cmd_diagnostic	},
	{	"error",		O_RDONLY,	cmd_error		},
//	{	"log",			O_RDONLY,	cmd_log			},
	{	"sense",		O_RDONLY,	cmd_sense		},
	{	"geometry",		O_RDONLY,	cmd_geometry	},
	{	"inquiry",		O_RDONLY,	cmd_inquiry		},
	{	"info",			O_RDONLY,	cmd_info		},
	{	0,				0,			0				}
};

void command(char *cmd) {
	static const char	delim[] = " \n\t=,:-";
	extern int      qnx_nidtostr(nid_t nid, char *bufp, int maxbuf);

	if(cmd[0] == '/') {
		if(fd != -1) {
			close(fd);
		}
		if(cmd[1] == '/') {
			strcpy(path, cmd);
		} else {
			sprintf(path, "//%ld%s", nid, cmd);
		}
		if(		(fd = open(path, oflag = O_RDWR)) == -1 &&
				(fd = open(path, oflag = O_RDONLY)) == -1) {
			fprintf(stderr, "Can't open device %s (%s)\n", path, strerror(errno));
		} else {
			printf("Opened device          : %s\n", path);
		}
		return;
	} else {
		int					fd2;
		int					oflag2;
		char				dev[NAME_MAX + 20];

		sprintf(dev, "//%ld/dev/%s", nid, cmd);
		if(		(fd2 = open(dev, oflag2 = O_RDWR)) != -1 ||
				(fd2 = open(dev, oflag2 = O_RDONLY)) != -1) {
			if(fd != -1) {
				close(fd);
			}
			fd = fd2;
			oflag = oflag2;
			strcpy(path, dev);
			printf("Opened device          : %s\n", path);
			return;
		} 
		if(errno != ENOENT) {
			fprintf(stderr, "Can't open device %s (%s)\n", dev, strerror(errno));
			return;
		}
	}
	for(cmd = strtok(cmd, delim); cmd; cmd = strtok(0, delim)) {
		struct cmds			*p1, *p;
		int					len;

		len = strlen(cmd);
		p = 0;
		for(p1 = cmds; p1->name; p1++) {
			if(!strnicmp(p1->name, cmd, len)) {
				if(p) {
					p = 0;
					break;
				} else {
					p = p1;
				}
			}
		}
		if(p) {
			if(fd == -1 && p->oflag != -1) {
				if(*path) {
					fprintf(stderr, "Couldn't open %s\n", path);
				} else {
					fprintf(stderr, "The %s command needs a device\n", cmd);
				}
				return;
			}
			if(p->oflag == O_RDWR && oflag != O_RDWR) {
				fprintf(stderr, "Couldn't open %s for write\n", path);
				return;
			}
			if(p->oflag == O_RDONLY && oflag != O_RDWR && oflag != O_RDONLY) {
				fprintf(stderr, "Couldn't open %s for read\n", path);
				return;
			}
			memset(&scsicmd, 0x00, sizeof scsicmd);
			scsicmd.uscsi_cdb = &scsicdb;
			memset(&scsicdb, 0x00, scsicmd.uscsi_cdblen = sizeof scsicdb);
			scsicmd.uscsi_bufaddr = &scsibuf;
			memset(&scsibuf, 0x00, scsicmd.uscsi_buflen = sizeof scsibuf);
			(p->func)();
		} else {
			fprintf(stderr, "Unknown command '%s'\n", cmd);
		}
	}
}

main(int argc, char *argv[]) {
	int			c;

	while((c = getopt(argc, argv, "n:")) !=-1) {
		switch(c) {
		case 'n':
			nid = qnx_strtonid(optarg, 0);
			break;
		}
	}
	if(optind >= argc) {
		command("scan");
	} else {
		while(optind < argc) {
			command(argv[optind++]);
		}
	}
	if(fd != -1) {
		close(fd);
	}
}
